"""This module contains the config model for the Wind plant."""
import pandas as pd

from pysimmods.model.config import ModelConfig
import pandas as pd


class TurbineConfig(ModelConfig):
    """Config parameters of Wind plant model.

    Parameters
    ----------
    params : dict
        Contains the configuration of the turbine. See attribute
        section for more information about the parameters, attributes
        marked with *(Input)* can or must be provided.

    Attributes
    ----------


    """

    def __init__(self, params):
        super().__init__(params)

        self.wind_speed_height = params.get("wind_speed_height", 11)
        self.roughness_length = params.get("roughness_length", 15)
        self.obstacle_height = params.get("obstacle_height", 5)
        self.air_density = params.get("air_density", 1.225)
        self.nominal_power = params.get("nominal_power", 3e6)
        self.hub_height = params.get("hub_height", 105)
        self.rotor_diameter = params.get("rotor_diameter", 60)
        self.turbine_type = params.get("turbine_type", "E-126/4200")
        self.pressure = params.get("pressure", 98405.7)
        self.temperature = params.get("temperature", 268)
        self.temperature_height = params.get("temperature_height", 10)
        power_curve_values = params.get(
            "power_curve_values",
            [
                p * 1000
                for p in [  # from kw to w
                    0.0,
                    26.0,
                    180.0,
                    1500.0,
                    3000.0,
                    3000.0,
                ]
            ],
        )
        power_curve_wind_speed = params.get(
            "power_curve_wind_speeds", [0.0, 2.0, 4.0, 6.0, 8.0, 10.0]
        )

        self.power_curve = pd.DataFrame(
            data={
                "value": power_curve_values,
                "wind_speed": power_curve_wind_speed,
            }
        )
        # print(self.power_curve)
        # self.power_curve = params.get(
        #    "power_curve",
        #    pd.DataFrame(
        #        data={
        #            "value": [
        #                p * 1000
        #                for p in [  # from kw to w
        #                    0.0,
        #                    26.0,
        #                    180.0,
        #                    1500.0,
        #                    3000.0,
        #                    3000.0,
        #                ]
        #            ],  # in W
        #            "wind_speed": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        #        }
        #    ),
        # )
        #     self.wind_speed = params.get(
        #        "wind_speed",
        #   pd.Series(data=[5.0, 6.5]),
        #    )

        # self.p_max_kw = self.eta_percent / 100 * self.a_m2
        self.p_min_kw = 0
        # self.max_step_size = 300
        self.default_p_schedule = None
        self.default_q_schedule = None



from pysimmods.generator.turbinesim.presets import turbine_preset


def wind_presets(
    pn_max_kw,
    cos_phi=0.9,
    q_control="prioritize_p",
    inverter_mode="inductive",
    **kwargs
):
    wparams, winit = turbine_preset(pn_max_kw, **kwargs)

    params = {
        "wind_turbine": wparams,
        "inverter": {
            "sn_kva": pn_max_kw / cos_phi,
            "q_control": q_control,
            "cos_phi": cos_phi,
            "inverter_mode": inverter_mode,
        },
        "sign_convention": "active",
    }

    inits = {"wind_turbine": winit, "inverter": None}

    return params, inits



import numpy as np
import pandas as pd


def logarithmic_profile(
    wind_speed,
    wind_speed_height,
    hub_height,
    roughness_length,
    obstacle_height=0.0,
):

    if (
        0.7 * obstacle_height > wind_speed_height
    ):  # not strong enough to overcome
        raise ValueError(
            "To take an obstacle height of {0} m ".format(obstacle_height)
            + "into consideration, wind "
            + "speed data of a greater height is needed."
        )
    # Return np.array if wind_speed is np.array       #  check if wind_speed is  a  np.array object
    if isinstance(wind_speed, np.ndarray) and isinstance(
        roughness_length, pd.Series
    ):
        roughness_length = np.array(roughness_length)

    return (
        wind_speed
        * np.log((hub_height - 0.7 * obstacle_height) / roughness_length)
        / np.log(
            (wind_speed_height - 0.7 * obstacle_height) / roughness_length
        )
    )


# the other method Hellme



"""This module contains a model of a wind turbine and an inverter"""
from ..turbinesim import WindPowerPlant
from .config import WindTurbineSystemConfig
from .inputs import WindTurbineSystemInputs
from .state import WindTurbineSystemState

from ...model.qgenerator import QGenerator
from ...other.invertersim.inverter import Inverter


class WindTurbineSystem(QGenerator):
    def __init__(self, params, inits):
        self.config = WindTurbineSystemConfig(params)
        self.inputs = WindTurbineSystemInputs()
        self.state = WindTurbineSystemState(inits)

        self.turbine = WindPowerPlant(
            params["wind_turbine"], inits["wind_turbine"]
        )
        self.inverter = Inverter(params["inverter"], inits["inverter"])

    def step(self):
        """Perform simulation step."""

        # Step the wind turbine
        self.turbine.inputs.step_size = self.inputs.step_size
        self.turbine.inputs.now_dt = self.inputs.now_dt
        self.turbine.inputs.wind_v_m_per_s = self.inputs.wind_v_m_per_s
        self.turbine.inputs.t_air_deg_celsius = self.inputs.t_air_deg_celsius
        self.turbine.step()

        # Step the inverter
        self.inverter.inputs.p_in_kw = self.turbine.state.p_kw
        self.inverter.inputs.p_set_kw = self.inputs.p_set_kw
        self.inverter.inputs.q_set_kvar = self.inputs.q_set_kvar
        self.inverter.inputs.cos_phi_set = self.inputs.cos_phi_set
        self.inverter.inputs.inductive = self.inputs.inverter_inductive
        self.inverter.step()

        # Update state
        self.state.p_kw = self.inverter.state.p_kw
        self.state.p_possible_max_kw = self.turbine.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state.inductive

        self.inputs.reset()

    def get_state(self):
        state_dict = {
            "wind_turbine": self.turbine.get_state(),
            "inverter": self.inverter.get_state(),
        }
        return state_dict

    def set_state(self, state_dict):
        self.turbine.set_state(state_dict["wind_turbine"])
        self.inverter.set_state(state_dict["inverter"])

        self.state.p_possible_max_kw = self.turbine.state.p_kw
        self.state.p_kw = self.inverter.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state.inductive

    def set_q_kvar(self, q_kvar: float):
        self.inputs.q_set_kvar = q_kvar



"""This module contains the input model for PV plant."""
from pysimmods.model.inputs import ModelInputs


class TurbineInputs(ModelInputs):
    """Input variables of Wind plant model.

    See :class:`pysimmods.model.inputs.ModelInputs` for additional
    information. This class has no inputs itself. Instead, each
    of the values is to be provided before each step.

    Attributes
    ----------
    t_air_deg_celsius:
        Air temperature in [°C].
    wind_v_m_per_s:
        Wind speed in meter per second
    """

    def __init__(self):
        super().__init__()
        self.t_air_deg_celsius = None
        self.wind_v_m_per_s = None



"""This module contains the *mosaik_api* meta definition for pysimmods
models.
"""
from pysimmods.buffer.batterysim import Battery
from pysimmods.consumer.hvacsim import HVAC
from pysimmods.generator.biogassim import BiogasPlant
from pysimmods.generator.chplpgsystemsim import CHPLPGSystem
from pysimmods.generator.dieselsim import DieselGenerator
from pysimmods.generator.pvsystemsim import PVPlantSystem
from pysimmods.generator.wind_turbine_system_sim import WindTurbineSystem

_SHARED_ATTRS = [
    "p_set_mw",
    "p_set_kw",
    "p_mw",
    "p_kw",
    "q_mvar",
    "q_kvar",
    "local_time",
]


META = {
    "type": "time-based",
    "models": {
        "Battery": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["soc_percent"] + _SHARED_ATTRS,
        },
        "Photovoltaic": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "q_set_mvar",
                "q_set_kvar",
                "t_air_deg_celsius",
                "t_module_deg_celsius",
                "bh_w_per_m2",
                "dh_w_per_m2",
                "s_module_w_per_m2",
                "inverter_inductive",
                "p_possible_max_mw",
            ]
            + _SHARED_ATTRS,
        },
        "CHP": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "day_avg_t_air_deg_celsius",
                "p_th_mw",
                "e_th_demand_set_kwh",
                "e_th_demand_kwh",
            ]
            + _SHARED_ATTRS,
        },
        "HVAC": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["t_air_deg_celsius"] + _SHARED_ATTRS,
        },
        "Biogas": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["p_th_mw"] + _SHARED_ATTRS,
        },
        "DieselGenerator": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": _SHARED_ATTRS,
        },
        "WindTurbine": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["wind_v_m_per_s", "t_air_deg_celsius"] + _SHARED_ATTRS,
        },
    },
}

MODELS = {
    "Photovoltaic": PVPlantSystem,
    "HVAC": HVAC,
    "Battery": Battery,
    "CHP": CHPLPGSystem,
    "DieselGenerator": DieselGenerator,
    "Biogas": BiogasPlant,
    "WindTurbine": WindTurbineSystem,
}



"""This module contains a Wind turbine."""

from copy import deepcopy

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pysimmods.generator.turbinesim.config import TurbineConfig
from pysimmods.generator.turbinesim.inputs import TurbineInputs
from pysimmods.generator.turbinesim.state import TurbineState
from pysimmods.model.generator import Generator


class WindPowerPlant(Generator):
    """Simulation model of a windturbine plant."""

    def __init__(self, params, inits):
        self.config = TurbineConfig(params)
        self.state = TurbineState(inits)
        self.inputs = TurbineInputs()

        # self.df = pd.DataFrame()

    def step(self):
        """Perform a simulation step."""
        next_state = deepcopy(self.state)
        self._check_inputs(next_state)

        # What is the formular? Where does it come from?
        wind_speed_hub = (
            self.inputs.wind_v_m_per_s
            * (self.config.hub_height / self.config.wind_speed_height) ** 1
            / 7
        )
        next_state.p_kw = np.interp(
            wind_speed_hub,
            self.config.power_curve["wind_speed"],
            self.config.power_curve["value"],
            left=0,
            right=0,
        )

        # Scale to step size
        next_state.p_kw *= self.inputs.step_size / 3600

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self, nstate):
        pass

    #
    #
    def temperature_correction(self):
        """Using air tempreture at hub height to correct the power out

        Linear temperature gradient model is used.

        (
            temperature_height (Air temperature in K.),
            temperature (Height in m for which the parameter
                `temperature` applies.
            )
        )
        Hub height of wind turbine in m.
        Temperature gradient of -6.5 K/km (-0.0065 K/m)

        """

        return (
            self.config.temperature,
            self.config.hub_height,
            self.config.temperature_height,
            self.config.temperature
            - 0.0065
            * (self.config.hub_height - self.config.temperature_height),
        )

    #  return self.config.temperature,self.config.hub_height,self.config.temperature_height

    # FIXME: Seems quite redundant right now
    def air_density_correction(self):
        """Formular for air density correction.

        Using air density at hub height to correct the power out.
        Describe the effect of air density-temperature on the generated
        power.
        """
        temperature_hub_height = self.temperature_correction()

        #        return (
        #     (self.config.pressure / 100 - (self.config.hub_height - self.config.pressure_height) * 1 / 8)
        #     * 100
        #     / (287.058 * temperature_hub_height)
        # )

        return temperature_hub_height

    def corr_poweroutput(self):
        out = np.array(self.config.power_curve["wind_speed"])

        #   return np.array(self.config.power_curve["wind_speed"]) , np.array(self.config.power_curve["value"] )

        #         self.config.power_curve["wind_speed"],
        #         self.config.power_curve["value"],

        power_curves_per_ts = (1.225 / self.config.air_density).reshape(
            -1, 1
        ) ** np.interp(
            self.config.power_curve["wind_speed"], [7.5, 12.5], [1 / 3, 2 / 3]
        )

        #                   * self.config.power_curve["wind_speed"]
        # ( (1.225 / self.config.air_density).reshape(-1, 1)
        #   ** np.interp(self.config.power_curve["wind_speed"], [7.5, 12.5], [1 / 3, 2 / 3])
        #   ) * self.config.power_curve["wind_speed"]
        return power_curves_per_ts

    #     # Create the interpolation function
    #     def interp_func(w_speed, p_curves):
    #         return np.interp(
    #         w_speed, p_curves, power_curve_values, left=0, right=0
    #         )

    #     # Calculate the power output by mapping the arrays to the interp function
    #     power_output = np.array(
    #     list(map(interp_func, wind_speed, power_curves_per_ts))
    #     )
    def barometric(
        pressure, pressure_height, hub_height, temperature_hub_height
    ):
        return (
            (pressure / 100 - (hub_height - pressure_height) * 1 / 8)
            * 1.225
            * 288.15
            * 100
            / (101330 * temperature_hub_height)
        )



"""This module contains the input model for the Wind Turbine System."""
from typing import Optional

from ...model.inputs import ModelInputs


class WindTurbineSystemInputs(ModelInputs):
    def __init__(self):
        super().__init__()

        self.cos_phi_set: Optional[float] = None
        self.inverter_inductive: Optional[bool] = None

        self.t_air_deg_celsius = None
        self.wind_v_m_per_s = None



"""This module contains the state model of the Wind Turbine System."""
from ...model.state import ModelState


class WindTurbineSystemState(ModelState):
    def __init__(self, inits):
        super().__init__(inits)

        self.p_possible_max_kw: float = 0
